------------------------------------------------------------------------------------
-- Bindings
------------------------------------------------------------------------------------

-- Binding Variables
BINDING_HEADER_SmartPack = "Smart Pack";
BINDING_NAME_SmartPackToggle = "Smart Pack Toggle";

------------------------------------------------------------------------------------
-- Main Frame
------------------------------------------------------------------------------------

SmP_ButtonsPerRow = 6;
SmP_VirtualBagWidth = SmP_ButtonsPerRow * 40 + 17;

function SmP_MainFrame_OnLoad(self)

	-- the main frame is an invisible top-level frame we use just to init some stuff
	
	if not SmP_PlusMinusDB then
		SmP_PlusMinusDB = {};
	end

	self:RegisterEvent("BAG_UPDATE");

end

function SmP_MainFrame_OnEvent(self, event, ...)

	if event == "BAG_UPDATE" and SmP_PrimaryBag:IsVisible() then
		SmP_UpdateVirtualBagsUI();
	end
	
end

------------------------------------------------------------------------------------
-- SmartPack
------------------------------------------------------------------------------------

function SmP_Toggle()

	if not SmP_PrimaryBag or not SmP_PrimaryBag:IsVisible() then
		PlaySound("igBackPackOpen");
		SmP_UpdateVirtualBagsUI(true);
		SmP_CloseAllBags();
	else
		PlaySound("igBackPackClose");
		for i,v in ipairs(SmP_VirtualBagsUI) do
			SmP_SetBagVisibility(v, false);
		end
	end
	
end

function SmP_CloseAllBags()
		
	CloseBackpack();
	
	for i=1,4 do
		CloseBag(i);
	end
		
end

------------------------------------------------------------------------------------
-- User Interface
------------------------------------------------------------------------------------

function SmP_UpdateVirtualBagsUI(open)

	SmP_UpdateVirtualBags();
	
	if not SmP_VirtualBagsUI then
		SmP_VirtualBagsUI = {};		
		table.insert(SmP_VirtualBagsUI, SmP_PrimaryBag);		
	end	

	for i,v in ipairs(SmP_VirtualBagsUI) do		
		SmP_SetBagVisibility(v, false);
	end

	local layoutController = { AtX = -CONTAINER_OFFSET_X - 5, AtY = CONTAINER_OFFSET_Y + 5, Column = 0, CanContinue = false };

	local slotCount, freeSlotCount = SmP_GetSlotInfo();
	SmP_PrimaryBag.SlotsPrompt:SetText(freeSlotCount.." of "..slotCount.." slots available.");

	SmP_PrimaryBag:SetWidth(SmP_VirtualBagWidth);
	SmP_SetBagLocation(SmP_PrimaryBag, layoutController.AtX, layoutController.AtY, open);
	SmP_SetBagHeight(SmP_PrimaryBag, 47, open);
	SmP_SetBagVisibility(SmP_PrimaryBag, true);

	layoutController.AtY = layoutController.AtY + 47 + 5;	
	
	for i,n in ipairs(SmP_VirtualBags.Names) do
		local continue = SmP_UpdateVirtualBagsUI_Bag(layoutController, n, open, nil, 0);
		local continueIndex = (continue and continue > 1) and 1 or 0;
		while continue do
			continue = SmP_UpdateVirtualBagsUI_Bag(layoutController, n, open, continue, continueIndex);
			continueIndex = continueIndex + 1;
		end
	end

end

function SmP_UpdateVirtualBagsUI_Bag(layoutController, bagName, open, continueFrom, continueIndex)

	local nameUI = "SmP_VirtualBagsUI_"..bagName..continueIndex;
	local currentBag = _G[nameUI];

	if not currentBag then
		
		currentBag = CreateFrame("Button", nameUI, nil, "SmP_BagTemplate");
		
		if continueFrom and continueFrom > 1 then
			currentBag.NamePrompt:SetText(bagName.." Cont.");
		else
			currentBag.NamePrompt:SetText(bagName);
		end
		
		currentBag.Sections = {};
		currentBag.NameUI = nameUI;
		currentBag.Name = bagName;
		
		table.insert(SmP_VirtualBagsUI, currentBag);
		
	end

	layoutController.InnerY = 7;
	
	for i,v in ipairs(currentBag.Sections) do
		v:Hide();
	end
	
	local continueRequired = nil;
	for i,n in ipairs(SmP_VirtualBags.Groups[bagName].SubNames) do
		if not continueFrom or i >= continueFrom then
			if not SmP_UpdateVirtualBagsUI_BagSection(currentBag, layoutController, n) then
				continueRequired = i;
				break;
			end
		end
	end

	
	local height = layoutController.InnerY + 25;
	currentBag:SetWidth(SmP_VirtualBagWidth);
	SmP_SetBagHeight(currentBag, height, open);
	
	if SmP_PlusMinusDB[bagName] then
		currentBag.MinusButton:Show();
		currentBag.PlusButton:Hide();
	else
		currentBag.PlusButton:Show();
		currentBag.MinusButton:Hide();
	end
	
	SmP_SetBagLocation(currentBag, layoutController.AtX, layoutController.AtY, open);
	layoutController.AtY = layoutController.AtY + height + 5;
	
	if continueRequired then
		layoutController.AtX = layoutController.AtX - SmP_VirtualBagWidth - 5;
		layoutController.AtY = CONTAINER_OFFSET_Y + 5;
		layoutController.Column = layoutController.Column + 1;
		layoutController.CanContinue = false;
	end
	
	if not continueRequired or continueRequired > 1 then	
		SmP_SetBagVisibility(currentBag, true);
	end
	
	return continueRequired;
	
end

function SmP_UpdateVirtualBagsUI_BagSection(bag, layoutController, sectionName)

	local nameUI = "SmP_VirtualBagsUI_"..bag.NameUI.."_"..sectionName;
	local currentSection = _G[nameUI];

	if not currentSection then
	
		currentSection = CreateFrame("Frame", nameUI, bag, "SmP_BagSectionTemplate");
		currentSection.NamePrompt:SetText(sectionName);
		currentSection.NameUI = nameUI;
		currentSection.Name = sectionName;
		currentSection.Buttons = {};
		
		table.insert(bag.Sections, currentSection);
		
	end
	
	local items = SmP_VirtualBags.Groups[bag.Name].SubGroups[sectionName];
	currentSection.Height = 20 + math.ceil(getn(items) / SmP_ButtonsPerRow) * 40;

	for i,v in ipairs(currentSection.Buttons) do
		v:Hide();
	end
	
	for i,v in ipairs(items) do
		SmP_UpdateVirtualBagsUI_BagSectionButton(bag, currentSection, i, v);
	end
	
	currentSection:SetHeight(currentSection.Height);
	currentSection:SetWidth(SmP_VirtualBagWidth);
	currentSection:SetPoint("BOTTOMLEFT", 
		currentSection:GetParent(), "BOTTOMLEFT", 
		10, layoutController.InnerY
	);
		
	local totalHeight = layoutController.AtY + layoutController.InnerY + currentSection.Height;
	if totalHeight > SmP_GetColumnHeight(layoutController.Column) and layoutController.CanContinue then
		return false;
	end
		
	layoutController.InnerY = layoutController.InnerY + currentSection.Height;
	layoutController.CanContinue = true;
	
	currentSection:Show();
	return true;
	
end

function SmP_UpdateVirtualBagsUI_BagSectionButton(bag, section, i, item)

	local nameUI = "SmP_VirtualBagsUI_"..bag.NameUI.."_"..section.NameUI.."_B"..i;
	local currentButton = section.Buttons[i];

	if not currentButton then
		currentButton = CreateFrame("Frame", nameUI, section, "SmP_BagButtonTemplate");		
		table.insert(section.Buttons, currentButton);		
	end

	local atX = bit.mod(i - 1, SmP_ButtonsPerRow) * 40;
	local atY = 20 + math.floor((i - 1) / SmP_ButtonsPerRow) * 40;
	
	currentButton:SetPoint("TOPLEFT", section, "TOPLEFT", atX, -atY);
	currentButton:SetID(item.BagId);
	currentButton.Button:SetID(item.SlotId);
	
	local texture, itemCount, locked, quality, readable = GetContainerItemInfo(item.BagId, item.SlotId);
		
	SetItemButtonTexture(currentButton.Button, texture);
	SetItemButtonCount(currentButton.Button, itemCount);
	SetItemButtonDesaturated(currentButton.Button, locked, 0.5, 0.5, 0.5);

	local r, g, b = GetItemQualityColor(item.Quality);	
	
	if BoG_TraderFrame and BoG_TraderFrame:IsVisible() 
	 and (not BoG_DB.Trader.Items[item.Name] or not BoG_DB.Trader.Items[item.Name].Enabled) then
		currentButton.Button.QualityTexture:SetVertexColor(1, 0, 0);
		currentButton.Button.QualityTexture:Show();	
	elseif item.Quality >= 2 then
		currentButton.Button.QualityTexture:SetVertexColor(r, g, b);
		currentButton.Button.QualityTexture:Show();
	else
		currentButton.Button.QualityTexture:Hide();
	end
	--currentButton.QualityTexture:Show();
	
	
	currentButton:Show();
	
end

function SmP_PlusMinus(button)

	local bag = button:GetParent();
	SmP_PlusMinusDB[bag.Name] = not SmP_PlusMinusDB[bag.Name];
	SmP_UpdateVirtualBagsUI();
	
end

function SmP_GetColumnHeight(column)
	if column == 0 then
		return GetScreenHeight() - CONTAINER_OFFSET_Y - 50;
	else
		return GetScreenHeight() - 30;
	end
end

function SmP_GetSlotInfo()

	local slotCount, freeSlotCount = 0, 0;

	for i=0,4 do
		slotCount = slotCount + GetContainerNumSlots(i);
		freeSlotCount = freeSlotCount + GetContainerNumFreeSlots(i);
	end
		
	return slotCount, freeSlotCount;
	
end

------------------------------------------------------------------------------------
-- DragDrop Support
------------------------------------------------------------------------------------

function SmP_AcceptItem(self, button)

	for i=0,4 do
		for j=1,GetContainerNumSlots(i) do
			local id = GetContainerItemLink(i,j);
			if not id then
				SmP_AcceptItemOnSlot(i, j, button);
				return;
			end
		end
	end

end

function SmP_AcceptItemOnSlot(bagId, slotId, button)
	
	local fakeSlot = { };
	local fakeParent = { };
	
	function fakeSlot:GetID()
		return slotId;
	end
	
	function fakeSlot:GetParent()
		return fakeParent;
	end
	
	function fakeParent:GetID()
		return bagId;
	end
	
	ContainerFrameItemButton_OnClick(fakeSlot, button);
	
end

------------------------------------------------------------------------------------
-- Cooldown/Lock Feedback
------------------------------------------------------------------------------------

function SmP_SetBagVisibility(bag, visible, immediate)

	if visible then
		bag.TargetAlpha = 1;
		bag:Show();
	else
		bag.TargetAlpha = 0;
	end

	if immediate then
		bag.CurrentAlpha = bag.TargetAlpha;
	end
	
end

function SmP_SetBagHeight(bag, height, immediate)
	
	bag.TargetHeight = height;
	
	if immediate then
		bag.CurrentHeight = height;
	elseif not bag:IsShown() then
		bag.CurrentHeight = 0;
	end
	
end

function SmP_SetBagLocation(bag, atX, atY, immediate)
	
	bag.TargetX = atX;
	bag.TargetY = atY;
	
	if immediate then
		bag.CurrentX = atX;
		bag.CurrentY = atY;
	elseif not bag:IsShown() and bag.TargetHeight then
		bag.CurrentX = atX;
		bag.CurrentY = atY - bag.TargetHeight;
	end
	
end

function SmP_UpdateBag(bag, elapsed)

	if bag.TargetX then
	
		bag.CurrentX = SmP_Animate2(bag.CurrentX or bag.TargetX, bag.TargetX, 0.3, elapsed);
		bag.CurrentY = SmP_Animate2(bag.CurrentY or bag.TargetY, bag.TargetY, 0.3, elapsed);

		bag:SetPoint("BOTTOMRIGHT", 
			bag:GetParent(), "BOTTOMRIGHT", 
			bag.CurrentX, bag.CurrentY
		);
	
	end	

	if bag.TargetHeight then
	
		bag.CurrentHeight = SmP_Animate2(bag.CurrentHeight or 0, bag.TargetHeight, 0.3, elapsed);		
		bag:SetHeight(bag.CurrentHeight);
		
	end
	
	if bag.TargetAlpha then
	
		bag.CurrentAlpha = SmP_Animate2(bag.CurrentAlpha or 0, bag.TargetAlpha, 0.3, elapsed);
		bag:SetAlpha(bag.CurrentAlpha);
		
		if bag.CurrentAlpha < 0.01 and bag.TargetAlpha == 0 then
			bag:Hide();
		end
		
	end

end

function SmP_UpdateSection(section)

	for i,v in ipairs(section.Buttons) do
		SmP_UpdateButton(v);
	end

end

function SmP_UpdateButton(button)
	
	local innerButton = button.Button;
	local bagId = button:GetID();
	local slotId = innerButton:GetID();
	
	SmP_UpdateButtonCooldown(innerButton, bagId, slotId);
	SmP_UpdateButtonLocked(innerButton, bagId, slotId);
	
end

function SmP_UpdateButtonCooldown(button, bagId, slotId)

	local cooldown = _G[button:GetName().."Cooldown"];
	local start, duration, enable = GetContainerItemCooldown(bagId, slotId);
	CooldownFrame_SetTimer(cooldown, start, duration, enable);
	if ( duration > 0 and enable == 0 ) then
		SetItemButtonTextureVertexColor(button, 0.4, 0.4, 0.4);
	else
		SetItemButtonTextureVertexColor(button, 1, 1, 1);
	end
	
end

function SmP_UpdateButtonLocked(button, bagId, slotId)
	
	local texture, itemCount, locked, quality, readable = GetContainerItemInfo(bagId, slotId);
	SetItemButtonDesaturated(button, locked, 0.5, 0.5, 0.5);
	
end

function SmP_Animate(current, target, rate, elapsed)

	if current < target then
		return math.min(current + rate * elapsed, target);
	elseif current > target then
		return math.max(current - rate * elapsed, target);
	end
	
	return current;	

end

function SmP_Animate2(current, target, rate, elapsed)

	if current < target then
		return math.min(current + (target - current) * rate, target);
	elseif current > target then
		return math.max(current + (target - current) * rate, target);
	end
	
	return current;	

end

------------------------------------------------------------------------------------
-- Sorter Engine
------------------------------------------------------------------------------------

function SmP_UpdateVirtualBags()

	SmP_VirtualBags = { Groups = {}, Names = {} };	

	for i=0,4 do
	
		local numSlots = GetContainerNumSlots(i);
		
		for j=1,numSlots do
			
			local item = GetContainerItemLink(i,j);
			local _,itemCount = GetContainerItemInfo(i,j);
			
			if item then
				
				local itemName,_,itemQuality,_,_,itemType,itemSubType,_,_,_,itemSellPrice = GetItemInfo(item);
				
				if not SmP_PlusMinusDB[itemType] then
					itemSubType = "All";
				end				
				
				if not SmP_VirtualBags.Groups[itemType] then
					table.insert(SmP_VirtualBags.Names, itemType);
					SmP_VirtualBags.Groups[itemType] = { SubGroups = {}, SubNames = {} };
				end
				
				if not SmP_VirtualBags.Groups[itemType].SubGroups[itemSubType] then
					table.insert(SmP_VirtualBags.Groups[itemType].SubNames, itemSubType);
					SmP_VirtualBags.Groups[itemType].SubGroups[itemSubType] = {};
				end
				
				table.insert(
					SmP_VirtualBags.Groups[itemType].SubGroups[itemSubType],
					{ 
						BagId = i, SlotId = j, 
						Name = itemName, 
						Quality = itemQuality, 
						SellPrice = itemSellPrice, 
						Count = itemCount 
					}
				);
				
			end
			
		end
		
	end
	
	-- Sort all the sections by name and all items by quality/value/count
	
	table.sort(SmP_VirtualBags.Names);	
	for k,v in pairs(SmP_VirtualBags.Groups) do
		table.sort(v.SubNames);
		for k2,v2 in pairs(v.SubGroups) do
			table.sort(v2, SmP_ItemSorter);
		end
	end
	
end

function SmP_ItemSorter(a, b)

	if a.Quality ~= b.Quality then
		return a.Quality > b.Quality;
	elseif a.SellPrice ~= b.SellPrice then
		return a.SellPrice > b.SellPrice;
	elseif a.Name ~= b.Name then
		return a.Name < b.Name;
	elseif a.Count ~= b.Count then
		return a.Count > b.Count;
	else
		return false;
	end

end